@@ -65,6 +65,7 @@
 #undef NO_FD_SET
 #undef NO_SYS_ERRLIST
 #undef NO_SYS_SIGLIST
+#undef NO_SYS_SIGLIST_DECL
 
 #undef WITH_SYMBOL_UNDERSCORE
 
@@ -91,6 +92,7 @@
 
 #undef G_THREAD_SOURCE
 
+#undef HAVE_GETPWUID_R_POSIX
 /* #undef PACKAGE */
 /* #undef VERSION */
 
